/*
  gpsview.cpp
  Copyright (C) 2008 Tomohiro Masubuchi <bucchi@e-ml.jp>

  This file is part of qgmap.

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; If not, see <http://www.gnu.org/licenses/>.
*/

#include "qgmap.h"
#include "gpsview.h"
#include "layeredbase.h"
#include "mainwindow.h"
#include "pos.h"
#include <qpixmap.h>
#include <qmessagebox.h>
#include <qpointarray.h>
#include <math.h>

static QPoint g_circle_tbl[360];

#define CIRCLE_RADIUS 20.0
#define PI 3.1415926535

#define DEG_NORMALIZE(deg) ((((int)(deg))+360)%360)
#define CPOINT(deg) g_circle_tbl[DEG_NORMALIZE(deg)]

GPSView::GPSView(LayeredBase *parent, const char *name) :
	LayeredWidget(parent, name)
{
	gps_on = false;
#ifdef QWS
	logger.setDeviceFile("/dev/rfcomm0");
#else
	logger.setDeviceFile("pipe");
#endif

	indicator = new GPSIndicator(parentWidget());
	indicator->hide();

	interval = 1;
	tid = -1;

	// create sin table
	for(int i = 0; i < 360; i++){
		int x, y;
		x = (int)(CIRCLE_RADIUS * sin(PI / 180.0 * (double)i));
		y = -(int)(CIRCLE_RADIUS * cos(PI / 180.0 * (double)i));
		CPOINT(i) = QPoint(x, y);
	}
	gps_pos_is_valid = false;
}

bool GPSView::paintEvent(QPaintEvent *e UNUSED, QPixmap &scr)
{
	if(!gps_on || !gps_pos_is_valid)
		return false;

	QPointArray poly(4);
	QPainter p(&scr);

	int deg = info.getDegree();

	QPoint c_rad(CIRCLE_RADIUS, CIRCLE_RADIUS);
	QRect geometry(scr_pos - c_rad, scr_pos + c_rad);

	if(!e->rect().intersects(geometry))
		return false;

	p.setPen(QPen(Qt::black));
	p.setBrush(QBrush(QColor(255,0,0)));

	poly.setPoint(0, scr_pos + CPOINT(deg));
	poly.setPoint(1, scr_pos + CPOINT(deg+180-35) * 4 / 5);
	poly.setPoint(2, scr_pos + CPOINT(deg+180) * 2 / 5);
	poly.setPoint(3, scr_pos + CPOINT(deg+180+35) * 4 / 5);
	p.drawPolygon(poly);

	p.drawLine(scr_pos + QPoint(-4, 0), scr_pos + QPoint(4, 0));
	p.drawLine(scr_pos + QPoint(0, -4), scr_pos + QPoint(0, 4));

#if 0
	printf("center.x/y = %d,%d\n", center.getX(), center.getY());
	printf("gps_pos.x/y = %d,%d\n", gps_pos.getX(), gps_pos.getY());
#endif
	return true;
}

bool GPSView::keyPressEvent(QKeyEvent *e)
{
	switch(e->key()){
	case Key_G:
		gps_on = !gps_on;
		parentWidget()->sendRequest(gps_on?"hide":"show", "freehand");
		if(gps_on){
			if(logger.start()){
				QMessageBox::critical(parentWidget(), "Error!!",
									  logger.getErrmsg());
				gps_on = false;
				parentWidget()->sendRequest("show", "freehand");

				gps_pos_is_valid = false;
				return true;
			}
			tid = startTimer(1000); // 1s timer
			indicator->setLock(true);
			indicator->show();
		}else{
			killTimer(tid);
			tid = -1;
			logger.stop();
			indicator->hide();
		}
		gps_pos_is_valid = false;
		repaint();
		return true;

	case Key_Space:
		if(!gps_on || !gps_pos_is_valid)
			break;

		indicator->setLock(true);
		emit detectGPSPosition(gps_pos);
		return true;
	}
	return false;
}

void GPSView::showEvent()
{
	QWidget *parent = parentWidget();
	width = parent->width();
	height = parent->height();

	setGeometry(0, 0, width, height);

	indicator->setGeometry(width - 5 - indicator->width(),
						   height - 5 - indicator->height(),
						   indicator->width(), indicator->height());
	scr_center = QPoint(width/2, height/2);
}

void GPSView::timerEvent(QTimerEvent *e)
{
//	printf("Timer\n");
	if(!logger.isLive()){
		QMessageBox::critical(parentWidget(), "Error!!",
							  logger.getErrmsg());
		gps_on = false;
		logger.stop();
		indicator->hide();
	}

	if(!logger.isStarted()){
		return;
	}
	killTimer(e->timerId());
	tid = -1;

	if(!gps_on){
		return;
	}

	tid = startTimer(interval * 1000); // timer

	info = logger.getInfo().copy();
	if(!info.isUpdated())
		return;

	if(datum == GpsDialog::Tokyo){
		info.convTokyoToWGS();
	}
	info.setTimeZone(timezone);

	indicator->updateInfo(info);

	if(info.isValid()){
		gps_pos = Pos(info.getLatitude(), info.getLongitude(), center.getZoom());
		gps_pos_is_valid = true;

		int center_x = width/2 + gps_pos.getX() - center.getX();
		int center_y = height/2 + gps_pos.getY() - center.getY();
		scr_pos = QPoint(center_x, center_y);
		repaint(center_x - CIRCLE_RADIUS, center_y - CIRCLE_RADIUS,
				CIRCLE_RADIUS * 2, CIRCLE_RADIUS * 2);

		if(indicator->isLock())
			emit detectGPSPosition(gps_pos);
	}
/*
	printf("time:%s lat:%lf lon:%lf\n",
		   info.getTime().toString().ascii(),
		   info.getLatitude(),
		   info.getLongitude());
*/
}

void GPSView::setPosition(Pos &pos)
{
	if(pos.getZoom() != gps_pos.getZoom()){
		gps_pos.zoomTo(pos.getZoom());
		repaint();
	} else if(pos != gps_pos){
		indicator->setLock(false);
	}

	center = pos;

	scr_pos = scr_center + gps_pos.getPoint() - center.getPoint();
}

void GPSView::setConfig(const QString &key, const QString &value)
{
	if(key=="GPSDev"){
		logger.setDeviceFile(value);
	} else if(key=="PreCommand"){
		logger.setInitCommand(value);
	} else if(key=="PostCommand"){
		logger.setEndCommand(value);
	} else if(key=="GPSInterval"){
		interval = value.toInt();
		if(tid >= 0){
			killTimer(tid);
			tid = startTimer(interval * 1000);
		}
		indicator->setInterval(interval*3);
	} else if(key=="GPSDatum"){
		datum = (GpsDialog::Datum)value.toInt();
	} else if(key=="GPSTimeZone"){
		timezone = value.toInt();
	}
}
